/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.constraint;

import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.createSimpleDescription;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterType;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getReturnType;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAbstract;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAssignableToCharSequence;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAssignableToCollection;
import javassist.CannotCompileException;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.constraints.NotEmpty;
import jp.objectfanatics.assertion.constraints.NotNull;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfParameterConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfReturnValueConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.impl.core.ConstraintAnnotationAssertionWeaver;
import jp.objectfanatics.assertion.weaver.impl.exception.IllegalUseOfParameterConstraintAnnotationExceptionImpl;
import jp.objectfanatics.assertion.weaver.impl.exception.IllegalUseOfReturnValueConstraintAnnotationExceptionImpl;
import jp.objectfanatics.commons.javassist.JavassistUtils;

/**
 * Implementation of {@link ConstraintAnnotationAssertionWeaver} for @{@link NotEmpty} annotation.
 * 
 * @author Makoto Sato
 */
public class NotEmptyAssertionWeaver extends AbstractAssertionWeaver {
	
	/**
	 * for asserting the value is not null.
	 */
	private NotNullAssertionWeaver notNullAssertionWeaver = new NotNullAssertionWeaver();
	
	/**
	 * Constructor.
	 */
	public NotEmptyAssertionWeaver() {
		super(NotEmpty.class);
	}
	
	@Override
	public void weaveParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfParameterConstraintAnnotationException {
		validateParameterConstraint(ctBehavior, paramIndex, annotation);
		doWeaveParameterConstraintAssertion(ctBehavior, paramIndex, annotation);
	}
	
	private void validateParameterConstraint(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfParameterConstraintAnnotationException {
		
		CtClass paramType = getParameterType(ctBehavior, paramIndex);
		
		// weaving is valid only for a type compatible with CharSequence, Collection, or array
		if (!isValidType(paramType)) {
			String description = createSimpleDescription(ctBehavior, paramType, annotation);
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(description, paramIndex, ctBehavior, annotation);
		}
	}
	
	private void doWeaveParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfParameterConstraintAnnotationException {
		CtClass paramType = getParameterType(ctBehavior, paramIndex);
		String paramValueIdentifier = JavassistUtils.toParamValueIdentifier(paramIndex);
		try {
			// weaves
			String valueName = "parameter " + paramIndex + "(zero base)";
			String assertionStatement = createAssertionStatement(ctBehavior, paramType, paramValueIdentifier, annotation, valueName, true);
			if (!isAbstract(ctBehavior)) ctBehavior.insertBefore(assertionStatement);
			
			// delegates @NotNull assertion
			notNullAssertionWeaver.weaveParameterConstraintAssertion(ctBehavior, paramIndex, annotation);
		}
		catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	@Override
	public void weaveReturnValueConstraintAssertion(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		CtClass returnType = getReturnType(ctMethod);
		validateReturnTypeConstraint(ctMethod, returnType, annotation);
		doWeaveReturnTypeConstraintAssertion(ctMethod, annotation);
	}
	
	// TODO: このメソッドって、iaValidTypeだけ異なるから、テンプレート化できるのでは？要検討。
	private void validateReturnTypeConstraint(CtMethod ctMethod, CtClass returnType, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		// weaving is valid only for a type compatible with CharSequence, Collection, or array
		if (!isValidType(returnType)) {
			String description = createSimpleDescription(ctMethod, returnType, annotation);
			throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(description, ctMethod, annotation);
		}
	}
	
	private void doWeaveReturnTypeConstraintAssertion(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		CtClass returnType = getReturnType(ctMethod);
		try {
			// delegates @NotNull assertion
			notNullAssertionWeaver.weaveReturnValueConstraintAssertion(ctMethod, annotation);
			
			// weaves
			String valueName = "return value";
			String assertionStatement = createAssertionStatement(ctMethod, returnType, "$_", annotation, valueName, false);
			if (!isAbstract(ctMethod)) ctMethod.insertAfter(assertionStatement); // TODO: insertBefore/After メソッドを、abstractなら無視する版を JavassistUtil内に作成しよう。
		}
		catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	@NotEmpty
	private String createAssertionStatement(@NotNull CtBehavior ctBehavior, @NotNull CtClass type, @NotEmpty String valueIdentifier, @NotNull Annotation annotation, @NotEmpty String valueName, boolean isParam) {
		String assertionErrorCondition;
		if (isAssignableToCharSequence(type)) {
			assertionErrorCondition = valueIdentifier + ".length() <= 0";
		}
		else if (type.isArray()) {
			assertionErrorCondition = valueIdentifier + ".length <= 0";
		}
		else if (isAssignableToCollection(type)) {
			assertionErrorCondition = valueIdentifier + ".size() <= 0";
		}
		else {
			throw new IllegalStateException("Unexpected type : " + type.getName());
		}
		String message = valueName + " is empty."; // TODO: use parameter name in the message.
		String assrtionStatement = isParam ? generateParameterAssrtionStatement(assertionErrorCondition, message) : generateReturnValueAssrtionStatement(assertionErrorCondition, message);
		return assrtionStatement;
	}
	
	/**
	 * Returns true if, and only if, the parameter type is compatible with CharSequence, Collection, or array.
	 * 
	 * @param paramType
	 *            target parameter type
	 * @return true if, and only if, the parameter type is compatible with CharSequence, Collection, or array.
	 */
	private boolean isValidType(@NotNull CtClass paramType) {
		return isAssignableToCharSequence(paramType) || paramType.isArray() || isAssignableToCollection(paramType);
	}
}
