/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.core;

import java.io.File;
import java.io.IOException;

import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.exception.NoWeavingNeededException;
import jp.objectfanatics.assertion.weaver.api.exception.NotClassFileException;

/**
 * {@link WeaverImpl} for class folder.
 * 
 * @author Makoto Sato
 */
public class ClassFolderWeaver {
	
	private IllegalUseOfConstraintAnnotationException firstIllegalUseOfConstraintAnnotationException = null;
	
	/**
	 * A {@link ClassFileWeaver} configured with classpaths.
	 */
	private final ClassFileWeaver classFileWeaver;
	
	/**
	 * Whether there are any illegal annotation definitions.
	 */
	private boolean hasIllegalConstraintAnnotationDefinition = false;
	
	/**
	 * Constructor.
	 * 
	 * @param classpaths
	 *            classpaths to be used to create {@link WeaverImpl} instance.
	 */
	public ClassFolderWeaver(String... classpaths) {
		this.classFileWeaver = new ClassFileWeaver(classpaths);
	}
	
	// TODO: 最初に受けたのを返すのではなく、全部まとめた例外を作成するのがたぶん正解。
	/**
	 * Weaves assertion functions into a class or class files in a class folder if needed.
	 * 
	 * @param classFolderOrClassFilePath
	 *            a path of a class folder or a class file
	 * @throws IOException
	 * @throws IllegalUseOfConstraintAnnotationException first IllegalUseOfConstraintAnnotationException
	 */
	public void weave(String classFolderOrClassFilePath, WeavingProgressListener listener) throws IOException, IllegalUseOfConstraintAnnotationException {
		weave(new File(classFolderOrClassFilePath), listener);
		if (hasIllegalConstraintAnnotationDefinition) throw firstIllegalUseOfConstraintAnnotationException;
	}
	
	private void weave(File classFolderOrClassFile, WeavingProgressListener listener) throws IOException {
		if (classFolderOrClassFile.isFile()) {
			try {
				classFileWeaver.weave(classFolderOrClassFile);
			}
			catch (IllegalUseOfConstraintAnnotationException e) {
				if (firstIllegalUseOfConstraintAnnotationException == null) firstIllegalUseOfConstraintAnnotationException = e;
				listener.notified(classFolderOrClassFile, e);
				hasIllegalConstraintAnnotationDefinition = true;
			}
			catch (NoWeavingNeededException e) {
				listener.notified(classFolderOrClassFile, e);
			}
			catch (NotClassFileException e) {
				listener.notified(classFolderOrClassFile, e);
			}
			return;
		}
		File[] files = classFolderOrClassFile.listFiles();
		if (files == null) return; // empty directory
		for (File child : files) {
			weave(child, listener);
		}
	}
	
	/**
	 * A listener interface for watching weaving progress.
	 */
	public interface WeavingProgressListener {
		void notified(File classFile, IllegalUseOfConstraintAnnotationException e);
		void notified(File classFile, NoWeavingNeededException e);
		void notified(File classFile, NotClassFileException e);
	}
}
