/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.exception;

import javassist.CtBehavior;
import javassist.CtConstructor;
import javassist.CtMethod;
import jp.objectfanatics.assertion.weaver.api.exception.LineNumberNotAvailableException;
import jp.objectfanatics.assertion.weaver.api.exception.info.AnnotationInfo;
import jp.objectfanatics.assertion.weaver.api.exception.info.BehaviorInfo;
import jp.objectfanatics.assertion.weaver.api.exception.info.ConstructorInfo;
import jp.objectfanatics.assertion.weaver.api.exception.info.MethodInfo;
import jp.objectfanatics.assertion.weaver.api.exception.info.ParameterInfo;
import jp.objectfanatics.assertion.weaver.api.exception.info.ParameterInfo.ParameterNameNotAvailableException;
import jp.objectfanatics.assertion.weaver.impl.exception.info.ConstructorInfoImpl;
import jp.objectfanatics.assertion.weaver.impl.exception.info.MethodInfoImpl;

//Checked 09.26.2009. Delete this line before release.
/**
 * A utility class for exceptions.
 * 
 * @author Makoto Sato
 */
public class ExceptionUtils {
	
	/**
	 * Returns specified behavior's location information string.
	 * <p/>
	 * ex. <code>at Hoge.foo(Hoge.java:17)</code><br/>
	 * ex. <code>at Foo.foo(Foo.java:-1)  // line number is not available</code><br/>
	 * ex. <code>at Hoge.<init>(Hoge.java:23)</code><br/>
	 * 
	 * @param behaviorInfo
	 *            {@link BehaviorInfo}
	 * @return specified behavior's location information string.
	 * @throws IllegalArgumentException
	 *             if <code>behaviorInfo</code> is <code>null</code>
	 */
	static String getLocationLineValue(BehaviorInfo behaviorInfo) {
		if (behaviorInfo == null) throw new IllegalArgumentException("behaviorInfo is null.");
		return "at " + getBehaviroNameBodyCompatibleWithPrintStackTrace(behaviorInfo) // ex. "at Hoge.<init>"
				+ "(" + getFileNameAndLineNumberEvenIfLineNumberIsNotAvailable(behaviorInfo) + ")" // ex. "(Hoge.java:-1)"
				+ getMessageForLineNumberNotAvailable(behaviorInfo) // ex. "  // line number is not available"
				+ "\n";
	}
	
	/**
	 * Returns a message for the case that the line number is not available.
	 * 
	 * @param behaviorInfo
	 *            {@link BehaviorInfo}
	 * @return a message for the case that the line number is not available
	 * @throws IllegalArgumentException
	 *             if <code>behaviorInfo</code> is <code>null</code>
	 */
	private static String getMessageForLineNumberNotAvailable(BehaviorInfo behaviorInfo) {
		if (behaviorInfo == null) throw new IllegalArgumentException("behaviorInfo is null.");
		try {
			behaviorInfo.getLineNumber();
			return "";
		}
		catch (LineNumberNotAvailableException e) {
			return "  // line number is not available";
		}
	}
	
	/**
	 * Returns file number and line number.
	 * 
	 * @param behaviorInfo
	 *            {@link BehaviorInfo}
	 * @return file number and line number
	 * @throws IllegalArgumentException
	 *             if <code>behaviorInfo</code> is <code>null</code>
	 */
	private static String getFileNameAndLineNumberEvenIfLineNumberIsNotAvailable(BehaviorInfo behaviorInfo) {
		if (behaviorInfo == null) throw new IllegalArgumentException("behaviorInfo is null.");
		String lineNumber;
		try {
			lineNumber = Integer.toString(behaviorInfo.getLineNumber());
		}
		catch (LineNumberNotAvailableException e) {
			lineNumber = "-1";
		}
		return behaviorInfo.getSourceFileName() + ":" + lineNumber;
	}
	
	/**
	 * Returns a behavior name compatible with printStackTrace().
	 * 
	 * @param behaviorInfo
	 *            {@link BehaviorInfo}
	 * @return behavior name without return type and parameters
	 * @throws IllegalArgumentException
	 *             if <code>behaviorInfo</code> is <code>null</code>
	 */
	private static String getBehaviroNameBodyCompatibleWithPrintStackTrace(BehaviorInfo behaviorInfo) {
		if (behaviorInfo == null) throw new IllegalArgumentException("behaviorInfo is null.");
		boolean isConstructor = ConstructorInfo.class.isInstance(behaviorInfo);
		return behaviorInfo.getLongName().split("\\(")[0] + (isConstructor ? ".<init>" : "");
	}
	
	/**
	 * Returns source file name and line number(line number will be attached only when it is available).
	 * 
	 * @param behaviorInfo
	 *            {@link BehaviorInfo}
	 * @return source file name and line number(line number will be attached only when it is available)
	 * @throws IllegalArgumentException
	 *             if <code>behaviorInfo</code> is <code>null</code>
	 */
	static String getSourceFileLineValue(BehaviorInfo behaviorInfo) {
		if (behaviorInfo == null) throw new IllegalArgumentException("behaviorInfo is null.");
		String lineNumber;
		try {
			lineNumber = ":" + Integer.toString(behaviorInfo.getLineNumber());
		}
		catch (LineNumberNotAvailableException e) {
			lineNumber = "";
		}
		return behaviorInfo.getSourceFileName() + lineNumber + "\n";
	}
	
	/**
	 * Returns {@link BehaviorInfo}.
	 * 
	 * @param ctBehavior
	 *            {@link CtBehavior}
	 * @return {@link BehaviorInfo}
	 * @throws if <code>ctBehavior</code> is <code>null</code>
	 */
	static BehaviorInfo createBehaviorInfo(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		if (CtConstructor.class.isInstance(ctBehavior)) {
			return new ConstructorInfoImpl(((CtConstructor) ctBehavior));
		}
		else {
			return new MethodInfoImpl(((CtMethod) ctBehavior));
		}
	}
	
	/**
	 * Returns a body of return type line for detailed description.
	 * 
	 * @param methodInfo
	 *            {@link MethodInfo}
	 * @return a body of return type line for detailed description
	 * @throws IllegalArgumentException
	 *             if <code>methodInfo</code> is <code>null</code>
	 */
	static String getReturnTypeLineBody(MethodInfo methodInfo) {
		if (methodInfo == null) throw new IllegalArgumentException("methodInfo is null.");
		return methodInfo.getReturnType().getName() + "\n";
	}
	
	/**
	 * Returns a body of method line for detailed description.
	 * 
	 * @param methodInfo
	 *            {@link MethodInfo}
	 * @return a body of method line for detailed description
	 * @throws IllegalArgumentException
	 *             if <code>methodInfo</code> is <code>null</code>
	 */
	static String getMethodLineBody(MethodInfo methodInfo) {
		if (methodInfo == null) throw new IllegalArgumentException("methodInfo is null.");
		return methodInfo.getReturnType().getSimpleName() + " " + methodInfo.getLongName() + "\n";
	}
	
	/**
	 * Returns a body of annotation line for detailed description.
	 * 
	 * @param annotationInfo
	 *            {@link AnnotationInfo}
	 * @return a body of annotation line for detailed description
	 * @throws IllegalArgumentException
	 *             if <code>annotationInfo</code> is <code>null</code>
	 */
	static String getAnnotationLineBody(AnnotationInfo annotationInfo) {
		if (annotationInfo == null) throw new IllegalArgumentException("annotationInfo is null.");
		return "@" + annotationInfo.getName() + "\n";
	}
	
	/**
	 * Returns a title of parameter line for detailed description.
	 * 
	 * @param parameterInfo
	 *            {@link ParameterInfo}
	 * @param headPad
	 *            length of the prefix padding
	 * @param tailPad
	 *            length of the suffix padding
	 * @return a title of parameter line for detailed description
	 * @throws IllegalArgumentException
	 *             if <code>parameterInfo</code> is <code>null</code>
	 */
	static String getParameterLineTitle(ParameterInfo parameterInfo, int headPad, int tailPad) {
		if (parameterInfo == null) throw new IllegalArgumentException("parameterInfo is null.");
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < headPad; i++)
			sb.append(" ");
		try {
			parameterInfo.getName();
			sb.append("parameter name ").toString();
		}
		catch (ParameterNameNotAvailableException e) {
			sb.append("parameter index").toString();
		}
		for (int i = 0; i < tailPad; i++)
			sb.append(" ");
		return sb.toString();
	}
	
	/**
	 * Returns a body of parameter line for detailed description.
	 * 
	 * @param annotationInfo
	 *            {@link AnnotationInfo}
	 * @return a body of parameter line for detailed description
	 * @throws IllegalArgumentException
	 *             if <code>parameterInfo</code> is <code>null</code>
	 */
	static String getParameterLineBody(ParameterInfo parameterInfo) {
		if (parameterInfo == null) throw new IllegalArgumentException("parameterInfo is null.");
		try {
			return parameterInfo.getName() + " [index = " + parameterInfo.getIndex() + "]\n";
		}
		catch (ParameterNameNotAvailableException e) {
			return parameterInfo.getIndex() + "\n";
		}
	}
	
	/**
	 * Returns a body of parameter type line for detailed description.
	 * 
	 * @param annotationInfo
	 *            {@link AnnotationInfo}
	 * @return a body of parameter type line for detailed description
	 * @throws IllegalArgumentException
	 *             if <code>parameterInfo</code> is <code>null</code>
	 */
	static String getParameterTypeLineBody(ParameterInfo parameterInfo) {
		if (parameterInfo == null) throw new IllegalArgumentException("parameterInfo is null.");
		return parameterInfo.getType().getName() + "\n";
	}
}
